#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages
from _version import __version__

setup(
    name='pytlgateway',
    version=__version__,
    description=(
        'scripts for ft'
    ),
    long_description=open('README.rst').read(),
    author='puyuan.tech',
    author_email='info@puyuan.tech',
    maintainer='puyuan_staff',
    maintainer_email='github@puyuan.tech',
    license='MIT License',
    packages=find_packages(),
    platforms=["all"],
    url='http://www.puyuan.tech',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'pymongo == 3.9.0'
        'requests'
        'websocket-client'
    ]
)